script_author('JustFedot')
script_name('[JF] Script Manager')
script_version('1.0.0')

local scriptregistry_url = "https://gitlab.com/public-samp-projects/script-manager/-/raw/master/scriptregistry.lua"

require("moonloader")
require ("sampfuncs")
local requests = require('requests')
local lfs = require("lfs")
local sampev = require("samp.events")
local effil = require("effil")
local encoding = require("encoding")
encoding.default = 'CP1251'
u8 = encoding.UTF8
local imgui = require("imgui")

function apply_custom_style()
    imgui.SwitchContext()
    local style = imgui.GetStyle()
    local colors = style.Colors
    local clr = imgui.Col
    local ImVec4 = imgui.ImVec4
    local ImVec2 = imgui.ImVec2
 
     style.WindowPadding = ImVec2(15, 15)
     style.WindowRounding = 15.0
     style.FramePadding = ImVec2(5, 5)
     style.ItemSpacing = ImVec2(12, 8)
     style.ItemInnerSpacing = ImVec2(8, 6)
     style.IndentSpacing = 25.0
     style.ScrollbarSize = 15.0
     style.ScrollbarRounding = 15.0
     style.GrabMinSize = 15.0
     style.GrabRounding = 7.0
     style.ChildWindowRounding = 8.0
     style.FrameRounding = 6.0
   
 
       colors[clr.Text] = ImVec4(0.95, 0.96, 0.98, 1.00)
       colors[clr.TextDisabled] = ImVec4(0.36, 0.42, 0.47, 1.00)
       colors[clr.WindowBg] = ImVec4(0.11, 0.15, 0.17, 1.00)
       colors[clr.ChildWindowBg] = ImVec4(0.15, 0.18, 0.22, 1.00)
       colors[clr.PopupBg] = ImVec4(0.08, 0.08, 0.08, 0.94)
       colors[clr.Border] = ImVec4(0.43, 0.43, 0.50, 0.50)
       colors[clr.BorderShadow] = ImVec4(0.00, 0.00, 0.00, 0.00)
       colors[clr.FrameBg] = ImVec4(0.20, 0.25, 0.29, 1.00)
       colors[clr.FrameBgHovered] = ImVec4(0.12, 0.20, 0.28, 1.00)
       colors[clr.FrameBgActive] = ImVec4(0.09, 0.12, 0.14, 1.00)
       colors[clr.TitleBg] = ImVec4(0.09, 0.12, 0.14, 0.65)
       colors[clr.TitleBgCollapsed] = ImVec4(0.00, 0.00, 0.00, 0.51)
       colors[clr.TitleBgActive] = ImVec4(0.08, 0.10, 0.12, 1.00)
       colors[clr.MenuBarBg] = ImVec4(0.15, 0.18, 0.22, 1.00)
       colors[clr.ScrollbarBg] = ImVec4(0.02, 0.02, 0.02, 0.39)
       colors[clr.ScrollbarGrab] = ImVec4(0.20, 0.25, 0.29, 1.00)
       colors[clr.ScrollbarGrabHovered] = ImVec4(0.18, 0.22, 0.25, 1.00)
       colors[clr.ScrollbarGrabActive] = ImVec4(0.09, 0.21, 0.31, 1.00)
       colors[clr.ComboBg] = ImVec4(0.20, 0.25, 0.29, 1.00)
       colors[clr.CheckMark] = ImVec4(0.28, 0.56, 1.00, 1.00)
       colors[clr.SliderGrab] = ImVec4(0.28, 0.56, 1.00, 1.00)
       colors[clr.SliderGrabActive] = ImVec4(0.37, 0.61, 1.00, 1.00)
       colors[clr.Button] = ImVec4(0.20, 0.25, 0.29, 1.00)
       colors[clr.ButtonHovered] = ImVec4(0.28, 0.56, 1.00, 1.00)
       colors[clr.ButtonActive] = ImVec4(0.06, 0.53, 0.98, 1.00)
       colors[clr.Header] = ImVec4(0.20, 0.25, 0.29, 0.55)
       colors[clr.HeaderHovered] = ImVec4(0.26, 0.59, 0.98, 0.80)
       colors[clr.HeaderActive] = ImVec4(0.26, 0.59, 0.98, 1.00)
       colors[clr.ResizeGrip] = ImVec4(0.26, 0.59, 0.98, 0.25)
       colors[clr.ResizeGripHovered] = ImVec4(0.26, 0.59, 0.98, 0.67)
       colors[clr.ResizeGripActive] = ImVec4(0.06, 0.05, 0.07, 1.00)
       colors[clr.CloseButton] = ImVec4(0.40, 0.39, 0.38, 0.16)
       colors[clr.CloseButtonHovered] = ImVec4(0.40, 0.39, 0.38, 0.39)
       colors[clr.CloseButtonActive] = ImVec4(0.40, 0.39, 0.38, 1.00)
       colors[clr.PlotLines] = ImVec4(0.61, 0.61, 0.61, 1.00)
       colors[clr.PlotLinesHovered] = ImVec4(1.00, 0.43, 0.35, 1.00)
       colors[clr.PlotHistogram] = ImVec4(0.90, 0.70, 0.00, 1.00)
       colors[clr.PlotHistogramHovered] = ImVec4(1.00, 0.60, 0.00, 1.00)
       colors[clr.TextSelectedBg] = ImVec4(0.25, 1.00, 0.00, 0.43)
       colors[clr.ModalWindowDarkening] = ImVec4(1.00, 0.98, 0.95, 0.73)
end
apply_custom_style()

do -- Xcfg Modified
    Xcfg = {
        _version    = 2.1,
        _author     = "Double Tap Inside",
        _modified   = "JustFedot",
        _email      = "double.tap.inside@gmail.com",
        _help = [[
            Module xcfg             = Xcfg()
                  Xcfg.
    
            nil                     = xcfg.mkpath(Str filename)
                  .
    
            Table loaded / nil      = xcfg.load(Str filename, [Bool save = false])
              .  'save'   true,     .
    
            Bool result             = xcfg.save(Str filename, Table new)
                .
    
            Bool result             = xcfg.insert(Str filename, (Value value or Int index), [Value value])
                .   ,   .
    
            Bool result             = xcfg.remove(Str filename, [Int index])
                .   ,    .
    
            Bool result             = xcfg.set(Str filename, (Int index or Str key), Value value)
                      .
    
            Bool result             = xcfg.update(Table old, (Table new or StrFilename new), [Bool overwrite = true])
                     . 'overwrite' ,    .
    
            Bool result             = xcfg.write(Str filename, Str str)
               ,   .
    
            Bool result             = xcfg.append(Str filename, Str str)
                .
    
            Table                   = xcfg.setupImcfg(Table cfg)
                   imgui     'cfg'.    ,   .
        ]]
    }
	function Xcfg.__init()
		local self = {}
		
		-- draw values
		local function draw_string(str)
			return string.format("%q", str)
		end
		
		local function is_var(key_or_index)
			if type(key_or_index) == "string" and key_or_index:match("^[_%a][_%a%d]*$") then
				return true
			
			else
				return false
			end
		end
		
		local function draw_table_key(key)
			if is_var(key) then
				return key
				
			else
				return "["..draw_key(key).."]"
			end
		end
		
		local function draw_table(tbl, tab)
			local tab = tab or ""
			local result = {}
			
			for key, value in pairs(tbl) do
				if type(value) == "string" then
					if type(key) == "number" and key <= #tbl then
						table.insert(result, draw_string(value))
						
					else
						table.insert(result, draw_table_key(key).." = "..draw_string(value))
					end
					
				elseif type(value) == "number" or type(value) == "boolean" then
					if type(key) == "number" and key <= #tbl then
						table.insert(result, tostring(value))
						
					else
						table.insert(result, draw_table_key(key).." = "..tostring(value))
					end
				
				elseif type(value) == "table" then
					if type(key) == "number" and key <= #tbl then
						table.insert(result, draw_table(value, tab.."\t"))
						
					else
						table.insert(result, draw_table_key(key).." = "..draw_table(value, tab.."\t"))
					end
					
				else
					if type(key) == "number" and key <= #tbl then
						table.insert(result, draw_string(tostring(value)))
						
					else
						table.insert(result, draw_table_key(key).." = "..draw_string(tostring(value)))
					end
				end
			end
			
			if #result == 0 and tab == "" then
				return ""
				
			elseif #result == 0 then
				return "{}"
			
			elseif tab == "" then
				return table.concat(result, ",\n")..",\n"
			
			else
				return "{\n"..tab..table.concat(result, ",\n"..tab)..",\n"..tab:sub(2).."}"
			end       
		end
		
		local function draw_value(value, tab)
			if type(value) == "string" then
				return draw_string(value)
			
			elseif type(value) == "number" or type(value) == "boolean" or type(value) == "nil" then
				return tostring(value)
			
			elseif type(value) == "table" then
				return draw_table(value, tab)
				
			else
				return draw_string(tostring(value))
			end
		end
		
		local function draw_key(key)
			if "string" == type(key) then
				return draw_string(key)
			
			elseif "number" == type(key) then
				return tostring(key)
			end
		end
		
		
		local function draw_config(tbl)
			local result = {}
		
			for key, value in pairs(tbl) do
				
				if type(key) == "number" then
					table.insert(result, "table.insert(tbl, "..draw_value(value, "\t")..")")
				
				elseif type(key) == "string" then			
					if is_var(key) then
						table.insert(result, "tbl."..draw_table_key(key).." = "..draw_value(value, "\t"))
					
					else
						table.insert(result, "tbl"..draw_table_key(key).." = "..draw_value(value, "\t"))
					end
				end
			end
			
			if #result == 0 then
				return ""
				
			else
				return table.concat(result, "\n").."\n"
			end
		end

		function self.load(filename, overwrite)
			assert(type(filename)=="string", ("bad argument #1 to 'load' (string expected, got %s)"):format(type(filename)))
			
			if overwrite == nil then
				overwrite = false
			end
			
			local file = io.open(filename, "r")
			
			if file then
				local text = file:read("*all")
				file:close()
				local lua_code = loadstring("local tbl = {}\n"..text.."\nreturn tbl")
				
				if lua_code then
					local result = lua_code()
					
					if type(result) == "table" then
						if overwrite then
							self.save(filename, result)
						end
						
						return result
					end
				end
			end
		end
		
		function self.save(filename, new)
			assert(type(filename)=="string", ("bad argument #1 to 'table_save' (string expected, got %s)"):format(type(filename)))
			assert(type(new)=="table", ("bad argument #2 to 'table_save' (table expected, got %s)"):format(type(new)))
		
			self.mkpath(filename)
			local file = io.open(filename, "w+")
			
			if file then
				local text = draw_config(new)
				file:write(text)
				file:close()
				
				return true
			else
				return false
			end
		end
		
		function self.insert(filename, value_or_index, value)
			assert(type(filename)=="string", ("bad argument #1 to 'insert' (string expected, got %s)"):format(type(filename)))
			
			if value then
				assert(type(value_or_index)=="number", ("bad argument #2 to 'insert' (number expected, got %s)"):format(type(value_or_index)))
			end
			
			local result
			
			if value then
				result = "table.insert(tbl, "..value_or_index..", "..draw_value(value, "\t")..")"
				
			else
				result = "table.insert(tbl, "..draw_value(value_or_index, "\t")..")"
			end
			
			self.mkpath(filename)
			local file = io.open(filename, "a+")
			
			if file then
				file:write(result.."\n")
				file:close()
				return true
				
			else
				return false
			end
		end
		
		function self.remove(filename, index)
			assert(type(filename)=="string", ("bad argument #1 to 'remove' (string expected, got %s)"):format(type(filename)))
			assert(type(index)=="number" or index == nil, ("bad argument #2 to 'remove' (number or nil expected, got %s)"):format(type(index)))
			local result
			
			if index then
				result = "table.remove(tbl, "..index..")"
				
			else
				result = "table.remove(tbl)"
			end
			
			self.mkpath(filename)
			local file = io.open(filename, "a+")
			
			if file then
				file:write(result.."\n")
				file:close()
				return true
				
			else
				return false
			end
		end
		
		function self.set(filename, key, value)
			assert(type(filename)=="string", ("bad argument #1 to 'set' (string expected, got %s)"):format(type(filename)))
			assert(type(key)=="number" or type(key)=="string", ("bad argument #2 to 'set' (number or string expected, got %s)"):format(type(key)))
			
			local result
					
			if is_var() then
				result = "tbl."..tostring(var).." = "..draw_value(value, "\t")
			
			else
				result = "tbl["..draw_key(key).."] = "..draw_value(value, "\t")
			end
			
			self.mkpath(filename)
			local file = io.open(filename, "a+")
			
			if file then
				file:write(result.."\n")
				file:close()
				return true
				
			else
				return false
			end
		end
		
		function self.mkpath(filename)
			assert(type(filename)=="string", ("bad argument #1 to 'mkpath' (string expected, got %s)"):format(type(filename)))
		
			local sep, pStr = package.config:sub(1, 1), ""
			local path = filename:match("(.+"..sep..").+$") or filename
			
			for dir in path:gmatch("[^" .. sep .. "]+") do
				pStr = pStr .. dir .. sep
				createDirectory(pStr)
			end
		end

		function self.update(old, new, overwrite)
			assert(type(old)=="table", ("bad argument #1 to 'update' (table expected, got %s)"):format(type(old)))
			assert(type(new)=="string" or type(new)=="table", ("bad argument #2 to 'update' (string or table expected, got %s)"):format(type(new)))
			
			if overwrite == nil then
				overwrite = true
			end
		
			if type(new) == "table" then
				if overwrite then
					for key, value in pairs(new) do
						old[key] = value
					end
					
				else
					for key, value in pairs(new) do
						if old[key] == nil then
							old[key] = value
						end
					end
				end
				
				return true
				
			elseif type(new) == "string" then
				local loaded = self.load(new)
				
				if loaded then
					if overwrite then
						for key, value in pairs(loaded) do
							old[key] = value
						end
						
					else
						for key, value in pairs(loaded) do
							if old[key] == nil then
								old[key] = value
							end
						end
					end
					
					return true
				end
			end
			
			return false
		end
		
		function self.append(filename, str)
			self.mkpath(filename)
			local file = io.open(filename, "a+")
			
			if file then
				file:write(str)
				file:close()
				return true
				
			else
				return false
			end
		end
		
		function self.write(filename, str)
			self.mkpath(filename)
			local file = io.open(filename, "w+")
			
			if file then
				file:write(str)
				file:close()
				return true
				
			else
				return false
			end
		end

        function self.setupImcfg(cfg)
            assert(type(cfg) == "table", ("bad argument #1 to 'setupImcfg' (table expected, got %s)"):format(type(cfg)))
            local function setupImcfgRecursive(cfg)
                local imcfg = {}
                for k, v in pairs(cfg) do
                    if type(v) == "table" then
                        imcfg[k] = setupImcfgRecursive(v)
                    elseif type(v) == "number" then
                        if v % 1 == 0 then
                            imcfg[k] = imgui.ImInt(v)
                        else
                            imcfg[k] = imgui.ImFloat(v)
                        end
                    elseif type(v) == "string" then
                        imcfg[k] = imgui.ImBuffer(256)
                        imcfg[k].v = u8(v)
                    elseif type(v) == "boolean" then
                        imcfg[k] = imgui.ImBool(v)
                    else
                        assert(false, ("Unsupported type for imcfg: %s"):format(type(v)))
                    end
                end
                return imcfg
            end
            return setupImcfgRecursive(cfg)
        end
		
		return self
	end
	setmetatable(Xcfg, {
	__call = function(self)
		return self.__init()
	end
})
end
local xcfg = Xcfg()

local filename = getWorkingDirectory()..'\\config\\'..thisScript().name..'\\config.cfg'

local imgui_windows = {
    main = imgui.ImBool(false),
}

local cfg = {
    lastUpdateTime = ".",
    savedList = {},
    installedScripts = {},
}
xcfg.update(cfg,filename)

function saveConfig()
    xcfg.save(filename,cfg)
end

local imcfg = xcfg.setupImcfg(cfg)

function addChat(a)
    sampAddChatMessage('{ffa500}'..thisScript().name..'{ffffff}: '..a, -1)
end

local doubleTapDebugTable = {}

function main()
    repeat wait(0) until isSampAvailable()
    while not isSampLoaded() do wait(0) end
    addChat('. : {ffc0cb}/jfm{ffffff}.')
    sampRegisterChatCommand('jfm',function()
        if imgui_windows.main.v then
            imgui_windows.main.v = false
        else
            local registry = getScriptRegistry()
            if registry then
                cfg.savedList = registry
                for k, v in pairs(cfg.savedList) do
                    v.description = trimMultilineString(v.description)
                end
                doubleTapDebugTable = simplifyTable(getListOfInstalledScripts())
                saveConfig()
                imgui_windows.main.v = true
            else
                addChat('')
            end
        end
    end)
    while true do wait(0)
        imgui.Process = imgui_windows.main.v
    end
end

function updateListOfInstalledScripts()
    doubleTapDebugTable = simplifyTable(getListOfInstalledScripts())
end

function simplifyTable(tbl)
    local path = getWorkingDirectory()..'\\'
    local res = {}
    for k,v in ipairs(tbl) do
        local file, err = io.open(path..v, 'r')
        if file then
            for line in file:lines() do
                local version = line:match("script_version%('(.+)'%)")
                if version then
                    res[v] = version
                    break
                end
            end
            file:close()
        else
            addChat('{ff0000}[] {ffffff}     .')
        end
    end
    return res
end

function enterpolyseFileName(str, reverse)
    if reverse then
        -- 
        return (str:gsub("%%(%x%x)", function(hex)
            return string.char(tonumber(hex, 16))
        end))
    else
        -- 
        return (str:gsub(".", function(char)
            return string.format("%%%02X", string.byte(char))
        end))
    end
end

function getListOfInstalledScripts()
    local path = getWorkingDirectory()
    local result = {}
    for filename in lfs.dir(path) do
        if filename:find('^%[JF%].+%.lua$') then
            table.insert(result,filename)
        end
    end
    return result
end


function getScriptCodeFromUrl(url)
    local response = requests.get(url)
    if response.status_code == 200 then
        return u8:decode(response.text)
    else
        return nil, ' : ' .. response.status_code
    end
end

function saveCodeAsFile(text, filename)
    local path = getWorkingDirectory()..'\\'..filename
    if not doesFileExist(path) then
        local file = io.open(path,'w')
        if file then
            file:write(text)
            file:close()
            return true
        else
            return false
        end
    end
end

function deleteScriptFile(filename)
    local path = getWorkingDirectory()..'\\'..filename
    if doesFileExist(path) then
        local success, err = os.remove(path)
        if success then
            addChat("{99ff99}[] {ffffff}  .")
        else
            addChat("{ff0000}[] {ffffff}   :")
            addChat(err)
        end
    else
        addChat('{ff0000}[] {ffffff}  .')
    end
end

function getScriptRegistry()
    local response = requests.get(scriptregistry_url)
    if response.status_code == 200 then
        local text = u8:decode(response.text)
        local zalupa = loadstring("return " .. text)
        return zalupa()
    else
        return nil, ' : ' .. response.status_code
    end
end

function createScriptFile(filename)
    local url = "https://gitlab.com/public-samp-projects/script-manager/-/raw/master/Scripts/"..enterpolyseFileName(filename, false)
    local code = getScriptCodeFromUrl(url)
    if code then
        if saveCodeAsFile(code, filename) then
            addChat('{99ff99}[] {ffffff} .  .')
        else
            addChat('{ff0000}[] {ffffff}     .')
        end
    else
        addChat('{ff0000}[] {ffffff}     .')
    end
end

local w,h = getScreenResolution()
local window_width,window_height = 800,600
local imTabs = 1
local updateName = ' '
function imgui.OnDrawFrame()

    if imgui_windows.main.v then


        imgui.SetNextWindowSize(imgui.ImVec2(window_width,window_height), imgui.Cond.FirstUseEver)
        imgui.SetNextWindowPos(imgui.ImVec2(w/2-window_width/2, h/2-window_height/2), imgui.Cond.FirstUseEver)
        imgui.Begin(u8(thisScript().name)..' Ver: '..thisScript().version.." ##main_window", imgui_windows.main, imgui.WindowFlags.NoCollapse + imgui.WindowFlags.NoScrollbar)

        imTabs = imgui.Tabs(2, imTabs, u8' ', u8(updateName))

        if updateName == ' ( )' then
            imgui.TextColoredRGB('{ff0000}!\n  .\n,    {ffff00}" " {ffffff} ,    {ffff00}" " {ffffff}.')
            if imgui.CenterButton(' ##updateAllButton', 200) then
                for k, v in pairs(cfg.savedList) do
                    if doubleTapDebugTable[v.name] ~= v.version then
                        deleteScriptFile(v.name)
                        createScriptFile(v.name)
                        updateListOfInstalledScripts()
                    end
                end
                updateName = ' '
            end
            imgui.Separator()
        end

        imgui.BeginChild('##mainChild', imgui.ImVec2(-1,-1), true)

            local buttonPos = imgui.GetWindowWidth()-imgui.GetStyle().ItemSpacing.x-80

            if imTabs==1 then
                for k, v in pairs(cfg.savedList) do
                    if not doubleTapDebugTable[v.name] and not v.isDeleted then
                        imgui.TextColoredRGB(': {FFFF00}'..v.name..'. {abcdef}('..v.realeseDate..')')
                        if imgui.IsItemHovered() then
                            imgui.SetTooltip(u8('    .'))
                        end
                        imgui.SameLine()
                        imgui.SetCursorPosX(buttonPos)
                        if imgui.Button(u8'##installScriptButton'..k, imgui.ImVec2(80,0)) then
                            createScriptFile(v.name)
                            updateListOfInstalledScripts()
                            updateName = ' '
                        end
                        imgui.TextDisabled(u8(v.short_description))
                        if imgui.CollapsingHeader(u8'  ##showDescButton'..k) then
                            imgui.TextColoredRGB(v.description)
                        end
                        imgui.Separator()
                    else
                        if not not v.isDeleted and doubleTapDebugTable[v.name] ~= v.version then
                            updateName = ' ( )'
                        end
                    end
                end

            elseif imTabs==2 then

                
                for k, v in pairs(cfg.savedList) do
                    if doubleTapDebugTable[v.name] then
                        if doubleTapDebugTable[v.name] ~= v.version then
                            updateName = ' ( )'
                        end
                        imgui.TextColoredRGB(': {FFFF00}'..v.name..'. {abcdef}('..v.realeseDate..')')
                        if imgui.IsItemHovered() then
                            imgui.SetTooltip(u8('    .'))
                        end
                        if v.isDeleted then
                            imgui.SameLine()
                            imgui.TextDisabled(u8("()"))
                            imgui.SameLine()
                            imgui.SetCursorPosX(buttonPos)
                            if imgui.Button(u8'##deleteScriptButton'..k, imgui.ImVec2(80,0)) then
                                deleteScriptFile(v.name)
                                updateListOfInstalledScripts()
                                updateName = ' '
                            end
                        else
                            imgui.SameLine()
                            if doubleTapDebugTable[v.name] ~= v.version then
                                imgui.TextColoredRGB('{99ff99} . {abcdef}('..doubleTapDebugTable[v.name]..' -> '..v.version..')')
                                imgui.SameLine()
                                imgui.SetCursorPosX(buttonPos-90)
                                if imgui.Button(u8'##UpdateScriptButton'..k, imgui.ImVec2(80,0)) then
                                    deleteScriptFile(v.name)
                                    createScriptFile(v.name)
                                    updateListOfInstalledScripts()
                                    updateName = ' '
                                end
                                imgui.SameLine()
                            end
                            imgui.SetCursorPosX(buttonPos)
                            if imgui.Button(u8'##deleteScriptButton'..k, imgui.ImVec2(80,0)) then
                                deleteScriptFile(v.name)
                                updateListOfInstalledScripts()
                                updateName = ' '
                            end
                            imgui.TextDisabled(u8(v.short_description))
                            if imgui.CollapsingHeader(u8'  ##showDescButton'..k) then
                                imgui.TextColoredRGB(v.description)
                            end
                        end
                        imgui.Separator()
                    end

                end

            end

        imgui.EndChild()

        imgui.End()

    end

end

function imgui.TextColoredRGB(text)
    local style = imgui.GetStyle()
    local colors = style.Colors
    local ImVec4 = imgui.ImVec4

    local explode_argb = function(argb)
        local a = bit.band(bit.rshift(argb, 24), 0xFF)
        local r = bit.band(bit.rshift(argb, 16), 0xFF)
        local g = bit.band(bit.rshift(argb, 8), 0xFF)
        local b = bit.band(argb, 0xFF)
        return a, r, g, b
    end

    local getcolor = function(color)
        if color:sub(1, 6):upper() == 'SSSSSS' then
            local r, g, b = colors[1].x, colors[1].y, colors[1].z
            local a = tonumber(color:sub(7, 8), 16) or colors[1].w * 255
            return ImVec4(r, g, b, a / 255)
        end
        local color = type(color) == 'string' and tonumber(color, 16) or color
        if type(color) ~= 'number' then return end
        local r, g, b, a = explode_argb(color)
        return imgui.ImColor(r, g, b, a):GetVec4()
    end

    local render_text = function(text_)
        for w in text_:gmatch('[^\r\n]+') do
            local text, colors_, m = {}, {}, 1
            w = w:gsub('{(......)}', '{%1FF}')
            while w:find('{........}') do
                local n, k = w:find('{........}')
                local color = getcolor(w:sub(n + 1, k - 1))
                if color then
                    text[#text], text[#text + 1] = w:sub(m, n - 1), w:sub(k + 1, #w)
                    colors_[#colors_ + 1] = color
                    m = n
                end
                w = w:sub(1, n - 1) .. w:sub(k + 1, #w)
            end
            if text[0] then
                for i = 0, #text do
                    imgui.TextColored(colors_[i] or colors[1], u8(text[i]))
                    imgui.SameLine(nil, 0)
                end
                imgui.NewLine()
            else imgui.Text(u8(w)) end
        end
    end

    render_text(text)
end

function imgui.Tabs(count, tabs, ...)
    local buttonLabels = {...}
    local style = imgui.GetStyle()
    local windowWidth = imgui.GetWindowWidth() - style.WindowPadding.x * 2
    local minButtonWidth = 0

    --    
    for i = 1, count do
        local labelWidth = imgui.CalcTextSize(buttonLabels[i] or "Button "..i).x + style.FramePadding.x * 2
        if labelWidth > minButtonWidth then
            minButtonWidth = labelWidth
        end
    end

    --         
    local buttonsPerRow = math.floor(windowWidth / minButtonWidth)
    if buttonsPerRow < 1 then buttonsPerRow = 1 end  -- ,        
    local rows = math.ceil(count / buttonsPerRow)

    for row = 1, rows do
        local buttonsInThisRow = math.min(buttonsPerRow, count - (row - 1) * buttonsPerRow)
        local buttonWidth = (windowWidth - (buttonsInThisRow - 1) * style.ItemSpacing.x) / buttonsInThisRow

        --    
        for i = 1, buttonsInThisRow do
            local buttonIndex = (row - 1) * buttonsPerRow + i
            local buttonStyle = imgui.GetStyle()
            local needStylePop = false

            if tabs == buttonIndex then
                imgui.PushStyleColor(imgui.Col.Button, buttonStyle.Colors[imgui.Col.CheckMark])
                imgui.PushStyleColor(imgui.Col.ButtonHovered, buttonStyle.Colors[imgui.Col.CheckMark])
                imgui.PushStyleColor(imgui.Col.ButtonActive, buttonStyle.Colors[imgui.Col.CheckMark])
                needStylePop = true
            end

            local buttonLabel = buttonLabels[buttonIndex] or "Button "..buttonIndex
            local buttonPressed = imgui.Button(buttonLabel, imgui.ImVec2(buttonWidth, 0))

            if needStylePop then
                imgui.PopStyleColor(3)
            end

            if buttonPressed and tabs ~= buttonIndex then
                tabs = buttonIndex
            end

            if i < buttonsInThisRow then
                imgui.SameLine()
            end
        end

        --    ,    
        -- if row < rows then
        --     imgui.NewLine()
        -- end
    end

    return tabs
end

function imgui.CenterButton(label, size)
    if size then
        imgui.SetCursorPosX((imgui.GetWindowWidth() - size) / 2)
        return imgui.Button(u8(label),imgui.ImVec2(size,0))
    else
        imgui.SetCursorPosX((imgui.GetWindowWidth() - imgui.CalcTextSize(u8(label)).x) / 2)
        return imgui.Button(u8(label))
    end
end

function trimMultilineString(s)
    local lines = {}
    for line in s:gmatch("[^\n]+") do
        table.insert(lines, line:match("^%s+(.+)") or ' ')
    end
    return table.concat(lines, "\n")
end
